@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2023 TileDB Inc.
+ * @copyright Copyright (c) 2017-2024 TileDB Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -693,7 +693,9 @@ TEST_CASE("C++ API: Consolidation of empty arrays", "[cppapi][consolidation]") {
 TEST_CASE(
     "C++ API: Consolidation of sequential fragment writes",
     "[cppapi][consolidation][sequential]") {
-  Context ctx;
+  tiledb::Config cfg;
+  cfg["sm.consolidation.total_buffer_size"] = "1048576";
+  Context ctx(cfg);
   VFS vfs(ctx);
   const std::string array_name = "cpp_unit_array";
 
@@ -743,6 +745,7 @@ TEST_CASE(
 TEST_CASE("C++ API: Encrypted array", "[cppapi][encryption]") {
   const char key[] = "0123456789abcdeF0123456789abcdeF";
   tiledb::Config cfg;
+  cfg["sm.consolidation.total_buffer_size"] = "1048576";
   cfg["sm.encryption_type"] = "AES_256_GCM";
   cfg["sm.encryption_key"] = key;
   Context ctx(cfg);
@@ -827,6 +830,7 @@ TEST_CASE("C++ API: Encrypted array, std::string key", "[cppapi][encryption]") {
   tiledb::Config cfg;
   cfg["sm.encryption_type"] = "AES_256_GCM";
   cfg["sm.encryption_key"] = key.c_str();
+  cfg["sm.consolidation.total_buffer_size"] = "1048576";
   Context ctx(cfg);
   VFS vfs(ctx);
   const std::string array_name = "cpp_unit_array";
@@ -1194,6 +1198,78 @@ TEST_CASE(
     vfs.remove_dir(array_name);
 }
 
+TEST_CASE(
+    "C++ API: Writing single byte cell with global order",
+    "[cppapi][std::byte]") {
+  bool serialize = false, refactored_query_v2 = false;
+  const std::string array_name = "cpp_unit_array";
+
+  Context ctx;
+  VFS vfs(ctx);
+  if (vfs.is_dir(array_name)) {
+    vfs.remove_dir(array_name);
+  }
+
+  auto datatype = GENERATE(
+      tiledb_datatype_t::TILEDB_BLOB,
+      tiledb_datatype_t::TILEDB_GEOM_WKB,
+      tiledb_datatype_t::TILEDB_GEOM_WKT);
+
+  // Create
+  Domain domain(ctx);
+  domain.add_dimension(Dimension::create<int>(ctx, "rows", {{0, 0}}, 1));
+  ArraySchema schema(ctx, TILEDB_DENSE);
+  schema.set_domain(domain).set_order({{TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR}});
+  schema.add_attribute(Attribute::create(ctx, "a", datatype));
+  Array::create(array_name, schema);
+
+  // Write
+  std::byte data_w{1};
+  Array array_w(ctx, array_name, TILEDB_WRITE);
+  Query query_w(ctx, array_w);
+  query_w.set_layout(TILEDB_GLOBAL_ORDER)
+      .set_data_buffer("a", (void*)(&data_w), 1);
+
+  SECTION("no serialization") {
+    serialize = false;
+  }
+#ifdef TILEDB_SERIALIZATION
+  SECTION("serialization enabled global order write") {
+    serialize = true;
+    refactored_query_v2 = GENERATE(true, false);
+  }
+#endif
+
+  // Submit query
+  test::ServerQueryBuffers server_buffers_;
+  auto rc = test::submit_query_wrapper(
+      ctx,
+      array_name,
+      &query_w,
+      server_buffers_,
+      serialize,
+      refactored_query_v2);
+  REQUIRE(rc == TILEDB_OK);
+  array_w.close();
+
+  // Read
+  Array array(ctx, array_name, TILEDB_READ);
+  Query query(ctx, array);
+  Subarray subarray(ctx, array);
+  subarray.add_range(0, 0, 0);
+  std::byte data;
+  query.set_layout(TILEDB_ROW_MAJOR)
+      .set_subarray(subarray)
+      .set_data_buffer("a", (void*)(&data), 1);
+  query.submit();
+  array.close();
+
+  REQUIRE(data == data_w);
+
+  if (vfs.is_dir(array_name))
+    vfs.remove_dir(array_name);
+}
+
 TEST_CASE("C++ API: Write cell with large cell val num", "[cppapi][sparse]") {
   const std::string array_name = "cpp_unit_array";
   Context ctx;
@@ -1879,7 +1955,7 @@ TEST_CASE("C++ API: Array write and read from MemFS", "[cppapi][memfs]") {
 TEST_CASE(
     "C++ API: Array on s3 with empty subfolders",
     "[cppapi][s3][empty_subfolders]") {
-  const std::string array_bucket = "s3://" + random_name("tiledb") + "/";
+  const std::string array_bucket = "s3://tiledb-" + random_label() + "/";
   const std::string array_name = array_bucket + "cpp_unit_array/";
 
   tiledb::Config cfg;
@@ -2084,3 +2160,86 @@ TEST_CASE(
   vfs.remove_dir(get_commit_dir(array_read.uri()));
   vfs.remove_dir(array_read.uri() + "/__schema");
 }
+
+TEST_CASE(
+    "C++ API: Close array with running query", "[cppapi][close-before-read]") {
+  const std::string array_name = "cpp_unit_array";
+  Context ctx;
+  VFS vfs(ctx);
+
+  if (vfs.is_dir(array_name))
+    vfs.remove_dir(array_name);
+
+  // Create
+  Domain domain(ctx);
+  domain.add_dimension(Dimension::create<int>(ctx, "rows", {{0, 3}}, 4))
+      .add_dimension(Dimension::create<int>(ctx, "cols", {{0, 3}}, 4));
+  ArraySchema schema(ctx, TILEDB_DENSE);
+  schema.set_domain(domain).set_order({{TILEDB_ROW_MAJOR, TILEDB_ROW_MAJOR}});
+  schema.add_attribute(Attribute::create<int>(ctx, "a"));
+  Array::create(array_name, schema);
+
+  // Write
+  std::vector<int> data_w = {
+      1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
+  Array array_w(ctx, array_name, TILEDB_WRITE);
+  Query query_w(ctx, array_w);
+  query_w.set_subarray(Subarray(ctx, array_w).set_subarray({0, 3, 0, 3}))
+      .set_layout(TILEDB_ROW_MAJOR)
+      .set_data_buffer("a", data_w);
+  query_w.submit();
+  array_w.close();
+
+  // Open for read.
+  Array array(ctx, array_name, TILEDB_READ);
+  std::vector<int> subarray_read = {0, 3, 0, 3};
+  std::vector<int> a_read(16);
+
+  Query query(ctx, array);
+  query.set_subarray(subarray_read)
+      .set_layout(TILEDB_ROW_MAJOR)
+      .set_data_buffer("a", a_read);
+  query.submit_async();
+  array.close();
+
+  uint64_t i = 0;
+  while (query.query_status() != Query::Status::COMPLETE) {
+    i++;
+  }
+
+  CHECK(i > 0);
+}
+
+TEST_CASE("C++ API: Read empty array", "[cppapi][read-empty-array]") {
+  const std::string array_name_1d = "cpp_unit_array_1d";
+  Context ctx;
+  VFS vfs(ctx);
+
+  bool dups = GENERATE(true, false);
+
+  if (vfs.is_dir(array_name_1d)) {
+    vfs.remove_dir(array_name_1d);
+  }
+
+  ArraySchema schema(ctx, TILEDB_SPARSE);
+  Domain domain(ctx);
+  domain.add_dimension(Dimension::create<int32_t>(ctx, "d", {{0, 1000}}, 1001));
+  schema.set_domain(domain);
+  schema.add_attribute(Attribute::create<int32_t>(ctx, "a"));
+  schema.set_allows_dups(dups);
+  Array::create(array_name_1d, schema);
+  Array array(ctx, array_name_1d, TILEDB_READ);
+
+  std::vector<int32_t> d(1);
+  std::vector<int32_t> a(1);
+  Query q(ctx, array, TILEDB_READ);
+  q.set_layout(TILEDB_UNORDERED);
+  q.set_data_buffer("d", d);
+  q.set_data_buffer("a", a);
+  q.submit();
+  array.close();
+
+  if (vfs.is_dir(array_name_1d)) {
+    vfs.remove_dir(array_name_1d);
+  }
+}
