@@ -31,6 +31,8 @@
  * load_tile_*_values and get_tile_* apis of fragment metadata.
  */
 
+#include <random>
+
 #include <test/support/tdb_catch.h>
 #include "test/support/src/helpers.h"
 #include "tiledb/sm/c_api/tiledb_struct_def.h"
@@ -746,12 +748,12 @@ struct CPPVarTileMetadataFx {
     std::default_random_engine random_engine;
 
     uint64_t max_string_size = 100;
-    uint64_t num_strings = 2000;
+    int num_strings = 2000;
 
     if (f == 0) {
       // Generate random, sorted strings for the string ascii type.
       strings_.reserve(num_strings);
-      for (uint64_t i = 0; i < num_strings; i++) {
+      for (int i = 0; i < num_strings; i++) {
         strings_.emplace_back(
             tiledb::test::random_string(rand() % max_string_size));
       }
@@ -1652,4 +1654,4 @@ TEST_CASE_METHOD(
   create_array();
   write_fragment();
   check_metadata();
-}
\ No newline at end of file
+}
