@@ -35,8 +35,10 @@
 
 #include <tiledb/common/logger_public.h>
 #include "test/support/src/coords_workaround.h"
+#include "test/support/src/mem_helpers.h"
 #include "tiledb.h"
 #include "tiledb/common/common.h"
+#include "tiledb/common/random/random_label.h"
 #include "tiledb/sm/array/array.h"
 #include "tiledb/sm/cpp_api/tiledb"
 #include "tiledb/sm/enums/layout.h"
@@ -50,6 +52,15 @@
 #include <string>
 #include <thread>
 
+/**
+ * Helper function to set environment variables across platforms.
+ *
+ * @param __name Name of the environment variable.
+ * @param __value Value of the environment variable.
+ * @return 0 on success, -1 on error.
+ */
+int setenv_local(const char* __name, const char* __value);
+
 // A mutex for protecting the thread-unsafe Catch2 macros.
 extern std::mutex catch2_macro_mutex;
 
@@ -480,14 +491,6 @@ void create_subarray(
     tiledb::Subarray** subarray,
     bool coalesce_ranges = false);
 
-/**
- * Helper method that creates a TileDB context and a VFS object.
- *
- * @param ctx The TileDB context to be created.
- * @param vfs The VFS object to be created.
- */
-void create_ctx_and_vfs(tiledb_ctx_t** ctx, tiledb_vfs_t** vfs);
-
 /**
  * Helper function to get the supported filesystems.
  * Supports VFS override via "--vfs" command line argument.
@@ -512,15 +515,6 @@ void get_supported_fs(
  */
 void open_array(tiledb_ctx_t* ctx, tiledb_array_t* array, tiledb_query_type_t);
 
-/**
- * Returns a random bucket name, with `prefix` as prefix and using
- * the thread id as a "random" suffix.
- *
- * @param prefix The prefix of the bucket name.
- * @return A random bucket name.
- */
-std::string random_name(const std::string& prefix);
-
 /**
  * Helper method that removes a directory.
  *
