@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2023 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2024 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -33,6 +33,7 @@
 #include "tiledb/common/common.h"
 
 #include "tiledb/common/logger.h"
+#include "tiledb/common/memory_tracker.h"
 #include "tiledb/sm/array/array.h"
 #include "tiledb/sm/array_schema/array_schema.h"
 #include "tiledb/sm/array_schema/array_schema_evolution.h"
@@ -59,8 +60,7 @@
 
 using namespace tiledb::common;
 
-namespace tiledb {
-namespace sm {
+namespace tiledb::sm {
 
 class ArrayException : public StatusException {
  public:
@@ -84,11 +84,8 @@ Array::Array(
     const URI& array_uri,
     StorageManager* storage_manager,
     ConsistencyController& cc)
-    : array_schema_latest_(nullptr)
-    , array_uri_(array_uri)
-    , array_dir_(storage_manager->resources(), array_uri)
+    : array_uri_(array_uri)
     , array_uri_serialized_(array_uri)
-    , encryption_key_(make_shared<EncryptionKey>(HERE()))
     , is_open_(false)
     , is_opening_or_closing_(false)
     , array_dir_timestamp_start_(0)
@@ -99,9 +96,7 @@ Array::Array(
     , resources_(storage_manager_->resources())
     , config_(resources_.config())
     , remote_(array_uri.is_tiledb())
-    , metadata_()
-    , metadata_loaded_(false)
-    , non_empty_domain_computed_(false)
+    , memory_tracker_(storage_manager->resources().create_memory_tracker())
     , consistency_controller_(cc)
     , consistency_sentry_(nullopt) {
 }
@@ -110,38 +105,16 @@ Array::Array(
 /*                API                */
 /* ********************************* */
 
-void Array::set_array_schema_latest(
-    const shared_ptr<ArraySchema>& array_schema) {
-  array_schema_latest_ = array_schema;
-}
-
-const ArraySchema& Array::array_schema_latest() const {
-  return *(array_schema_latest_.get());
-}
-
-shared_ptr<const ArraySchema> Array::array_schema_latest_ptr() const {
-  return array_schema_latest_;
-}
-
-void Array::set_array_schemas_all(
-    std::unordered_map<std::string, shared_ptr<ArraySchema>>& all_schemas) {
-  array_schemas_all_ = all_schemas;
-}
-
 const URI& Array::array_uri() const {
   return array_uri_;
 }
 
-const ArrayDirectory& Array::array_directory() const {
-  return array_dir_;
-}
-
 const URI& Array::array_uri_serialized() const {
   return array_uri_serialized_;
 }
 
 const EncryptionKey* Array::encryption_key() const {
-  return encryption_key_.get();
+  return opened_array_->encryption_key();
 }
 
 // Used in Consolidator
@@ -162,25 +135,28 @@ Status Array::open_without_fragments(
                           "remote arrays are not supported."));
   }
 
-  metadata_.clear();
-  metadata_loaded_ = false;
-  non_empty_domain_computed_ = false;
+  opened_array_ = make_shared<OpenedArray>(
+      HERE(),
+      resources_,
+      memory_tracker_,
+      array_uri_,
+      encryption_type,
+      encryption_key,
+      key_length,
+      timestamp_start(),
+      timestamp_end_opened_at(),
+      is_remote());
 
   /* Note: query_type_ MUST be set before calling set_array_open()
     because it will be examined by the ConsistencyController. */
   query_type_ = QueryType::READ;
+  memory_tracker_->set_type(MemoryTrackerType::ARRAY_READ);
 
   /* Note: the open status MUST be exception safe. If anything interrupts the
    * opening process, it will throw and the array will be set as closed. */
   try {
     set_array_open(query_type_);
 
-    // Copy the key bytes.
-    st = encryption_key_->set_key(encryption_type, encryption_key, key_length);
-    if (!st.ok()) {
-      throw StatusException(st);
-    }
-
     if (remote_) {
       auto rest_client = resources_.rest_client();
       if (rest_client == nullptr) {
@@ -195,7 +171,7 @@ Status Array::open_without_fragments(
         if (!st.ok()) {
           throw StatusException(st);
         }
-        array_schema_latest_ = array_schema_latest.value();
+        set_array_schema_latest(array_schema_latest.value());
       } else {
         auto st = rest_client->post_array_from_rest(
             array_uri_, storage_manager_, this);
@@ -207,11 +183,13 @@ Status Array::open_without_fragments(
       {
         auto timer_se = resources_.stats().start_timer(
             "array_open_without_fragments_load_directory");
-        array_dir_ = ArrayDirectory(
-            resources_, array_uri_, 0, UINT64_MAX, ArrayDirectoryMode::READ);
+        set_array_directory(ArrayDirectory(
+            resources_, array_uri_, 0, UINT64_MAX, ArrayDirectoryMode::READ));
       }
-      std::tie(array_schema_latest_, array_schemas_all_) =
+      auto&& [array_schema_latest, array_schemas_all] =
           open_for_reads_without_fragments();
+      set_array_schema_latest(array_schema_latest);
+      set_array_schemas_all(std::move(array_schemas_all));
     }
   } catch (...) {
     set_array_closed();
@@ -229,14 +207,14 @@ void Array::load_fragments(
 
   // Load the fragment metadata
   std::unordered_map<std::string, std::pair<Tile*, uint64_t>> offsets;
-  fragment_metadata_ = FragmentMetadata::load(
+  set_fragment_metadata(FragmentMetadata::load(
       resources_,
       memory_tracker(),
-      array_schema_latest_ptr(),
+      opened_array_->array_schema_latest_ptr(),
       array_schemas_all(),
       *encryption_key(),
       fragments_to_load,
-      offsets);
+      offsets));
 }
 
 Status Array::open(
@@ -269,10 +247,12 @@ Status Array::open(
         Status_ArrayError("Cannot open array; Array already open."));
   }
 
-  metadata_.clear();
-  metadata_loaded_ = false;
-  non_empty_domain_computed_ = false;
   query_type_ = query_type;
+  if (query_type_ == QueryType::READ) {
+    memory_tracker_->set_type(MemoryTrackerType::ARRAY_READ);
+  } else {
+    memory_tracker_->set_type(MemoryTrackerType::ARRAY_WRITE);
+  }
 
   set_timestamps(
       timestamp_start, timestamp_end, query_type_ == QueryType::READ);
@@ -334,11 +314,17 @@ Status Array::open(
           "Cannot open array; encrypted remote arrays are not supported.");
     }
 
-    // Copy the key bytes.
-    st = encryption_key_->set_key(encryption_type, encryption_key, key_length);
-    if (!st.ok()) {
-      throw StatusException(st);
-    }
+    opened_array_ = make_shared<OpenedArray>(
+        HERE(),
+        resources_,
+        memory_tracker_,
+        array_uri_,
+        encryption_type,
+        encryption_key,
+        key_length,
+        this->timestamp_start(),
+        timestamp_end_opened_at(),
+        is_remote());
 
     if (remote_) {
       auto rest_client = resources_.rest_client();
@@ -350,7 +336,7 @@ Status Array::open(
         auto&& [st, array_schema_latest] =
             rest_client->get_array_schema_from_rest(array_uri_);
         throw_if_not_ok(st);
-        array_schema_latest_ = array_schema_latest.value();
+        set_array_schema_latest(array_schema_latest.value());
       } else {
         auto st = rest_client->post_array_from_rest(
             array_uri_, storage_manager_, this);
@@ -362,57 +348,59 @@ Status Array::open(
       {
         auto timer_se =
             resources_.stats().start_timer("array_open_read_load_directory");
-        array_dir_ = ArrayDirectory(
+        set_array_directory(ArrayDirectory(
             resources_,
             array_uri_,
             array_dir_timestamp_start_,
-            array_dir_timestamp_end_);
+            array_dir_timestamp_end_));
       }
-      std::tie(array_schema_latest_, array_schemas_all_, fragment_metadata_) =
+      auto&& [array_schema_latest, array_schemas_all, fragment_metadata] =
           open_for_reads();
+      set_array_schema_latest(array_schema_latest);
+      set_array_schemas_all(std::move(array_schemas_all));
+      set_fragment_metadata(std::move(fragment_metadata));
     } else if (
         query_type == QueryType::WRITE ||
         query_type == QueryType::MODIFY_EXCLUSIVE) {
       {
         auto timer_se =
             resources_.stats().start_timer("array_open_write_load_directory");
-        array_dir_ = ArrayDirectory(
+        set_array_directory(ArrayDirectory(
             resources_,
             array_uri_,
             array_dir_timestamp_start_,
             array_dir_timestamp_end_,
-            ArrayDirectoryMode::SCHEMA_ONLY);
+            ArrayDirectoryMode::SCHEMA_ONLY));
       }
       auto&& [st, array_schema_latest, array_schemas] = open_for_writes();
       throw_if_not_ok(st);
 
       // Set schemas
-      array_schema_latest_ = array_schema_latest.value();
-      array_schemas_all_ = array_schemas.value();
+      set_array_schema_latest(array_schema_latest.value());
+      set_array_schemas_all(std::move(array_schemas.value()));
 
       // Set the timestamp
-      metadata_.reset(timestamp_for_new_component());
-
+      opened_array_->metadata().reset(timestamp_for_new_component());
     } else if (
         query_type == QueryType::DELETE || query_type == QueryType::UPDATE) {
       {
         auto timer_se = resources_.stats().start_timer(
             "array_open_delete_or_update_load_directory");
-        array_dir_ = ArrayDirectory(
+        set_array_directory(ArrayDirectory(
             resources_,
             array_uri_,
             array_dir_timestamp_start_,
             array_dir_timestamp_end_,
-            ArrayDirectoryMode::READ);
+            ArrayDirectoryMode::READ));
       }
-      auto&& [st, array_schema_latest, array_schemas] = open_for_writes();
+      auto&& [st, latest, array_schemas] = open_for_writes();
       throw_if_not_ok(st);
 
       // Set schemas
-      array_schema_latest_ = array_schema_latest.value();
-      array_schemas_all_ = array_schemas.value();
+      set_array_schema_latest(latest.value());
+      set_array_schemas_all(std::move(array_schemas.value()));
 
-      auto version = array_schema_latest_->version();
+      auto version = array_schema_latest().version();
       if (query_type == QueryType::DELETE &&
           version < constants::deletes_min_version) {
         std::stringstream err;
@@ -433,7 +421,7 @@ Status Array::open(
       }
 
       // Updates the timestamp to use for metadata.
-      metadata_.reset(timestamp_for_new_component());
+      opened_array_->metadata().reset(timestamp_for_new_component());
     } else {
       throw ArrayException("Cannot open array; Unsupported query type.");
     }
@@ -455,10 +443,7 @@ Status Array::close() {
     return Status::Ok();
   }
 
-  non_empty_domain_.clear();
-  non_empty_domain_computed_ = false;
   clear_last_max_buffer_sizes();
-  fragment_metadata_.clear();
 
   try {
     set_array_closed();
@@ -468,10 +453,10 @@ Status Array::close() {
       // user
       if ((query_type_ == QueryType::WRITE ||
            query_type_ == QueryType::MODIFY_EXCLUSIVE) &&
-          metadata_.num() > 0) {
+          opened_array_->metadata().num() > 0) {
         // Set metadata loaded to be true so when serialization fetchs the
         // metadata it won't trigger a deadlock
-        metadata_loaded_ = true;
+        set_metadata_loaded(true);
         auto rest_client = resources_.rest_client();
         if (rest_client == nullptr) {
           throw Status_ArrayError(
@@ -483,15 +468,11 @@ Status Array::close() {
             array_dir_timestamp_end_,
             this));
       }
-
-      // Storage manager does not own the array schema for remote arrays.
-      array_schema_latest_.reset();
     } else {
-      array_schema_latest_.reset();
       if (query_type_ == QueryType::WRITE ||
           query_type_ == QueryType::MODIFY_EXCLUSIVE) {
         st = storage_manager_->store_metadata(
-            array_uri_, *encryption_key_.get(), &metadata_);
+            array_uri_, *encryption_key(), &opened_array_->metadata());
         if (!st.ok()) {
           throw StatusException(st);
         }
@@ -502,9 +483,7 @@ Status Array::close() {
       }
     }
 
-    array_schemas_all_.clear();
-    metadata_.clear();
-    metadata_loaded_ = false;
+    opened_array_.reset();
   } catch (std::exception& e) {
     is_opening_or_closing_ = false;
     throw Status_ArrayError(e.what());
@@ -587,7 +566,7 @@ std::vector<shared_ptr<const Enumeration>> Array::get_enumerations(
   // Dedupe requested names and filter out anything already loaded.
   std::unordered_set<std::string> enmrs_to_load;
   for (auto& enmr_name : enumeration_names) {
-    if (array_schema_latest_->is_enumeration_loaded(enmr_name)) {
+    if (array_schema_latest().is_enumeration_loaded(enmr_name)) {
       continue;
     }
     enmrs_to_load.insert(enmr_name);
@@ -615,30 +594,31 @@ std::vector<shared_ptr<const Enumeration>> Array::get_enumerations(
           array_dir_timestamp_start_,
           array_dir_timestamp_end_,
           this,
-          names_to_load);
+          names_to_load,
+          memory_tracker_);
     } else {
       // Create a vector of paths to be loaded.
       std::vector<std::string> paths_to_load;
       for (auto& enmr_name : enmrs_to_load) {
-        auto path = array_schema_latest_->get_enumeration_path_name(enmr_name);
+        auto path = array_schema_latest().get_enumeration_path_name(enmr_name);
         paths_to_load.push_back(path);
       }
 
       // Load the enumerations from storage
-      loaded = array_dir_.load_enumerations_from_paths(
-          paths_to_load, get_encryption_key(), memory_tracker_);
+      loaded = array_directory().load_enumerations_from_paths(
+          paths_to_load, *encryption_key(), memory_tracker_);
     }
 
     // Store the loaded enumerations in the schema
     for (auto& enmr : loaded) {
-      array_schema_latest_->store_enumeration(enmr);
+      opened_array_->array_schema_latest_ptr()->store_enumeration(enmr);
     }
   }
 
   // Return the requested list of enumerations
   std::vector<shared_ptr<const Enumeration>> ret(enumeration_names.size());
   for (size_t i = 0; i < enumeration_names.size(); i++) {
-    ret[i] = array_schema_latest_->get_enumeration(enumeration_names[i]);
+    ret[i] = array_schema_latest().get_enumeration(enumeration_names[i]);
   }
   return ret;
 }
@@ -648,11 +628,11 @@ void Array::load_all_enumerations() {
     throw ArrayException("Unable to load all enumerations; Array is not open.");
   }
   // Load all enumerations, discarding the returned list of loaded enumerations.
-  get_enumerations(array_schema_latest_->get_enumeration_names());
+  get_enumerations(array_schema_latest().get_enumeration_names());
 }
 
 bool Array::is_empty() const {
-  return fragment_metadata_.empty();
+  return opened_array_->fragment_metadata().empty();
 }
 
 bool Array::is_open() {
@@ -664,10 +644,6 @@ bool Array::is_remote() const {
   return remote_;
 }
 
-std::vector<shared_ptr<FragmentMetadata>> Array::fragment_metadata() const {
-  return fragment_metadata_;
-}
-
 tuple<Status, optional<shared_ptr<ArraySchema>>> Array::get_array_schema()
     const {
   // Error if the array is not open
@@ -677,7 +653,7 @@ tuple<Status, optional<shared_ptr<ArraySchema>>> Array::get_array_schema()
             Status_ArrayError("Cannot get array schema; Array is not open")),
         nullopt};
 
-  return {Status::Ok(), array_schema_latest_};
+  return {Status::Ok(), opened_array_->array_schema_latest_ptr()};
 }
 
 QueryType Array::get_query_type() const {
@@ -706,22 +682,22 @@ Status Array::get_max_buffer_size(
   }
 
   // Not applicable to heterogeneous domains
-  if (!array_schema_latest_->domain().all_dims_same_type()) {
+  if (!array_schema_latest().domain().all_dims_same_type()) {
     return LOG_STATUS(
         Status_ArrayError("Cannot get max buffer size; Function not "
                           "applicable to heterogeneous domains"));
   }
 
   // Not applicable to variable-sized dimensions
-  if (!array_schema_latest_->domain().all_dims_fixed()) {
+  if (!array_schema_latest().domain().all_dims_fixed()) {
     return LOG_STATUS(Status_ArrayError(
         "Cannot get max buffer size; Function not "
         "applicable to domains with variable-sized dimensions"));
   }
 
   // Check if name is attribute or dimension
-  bool is_dim = array_schema_latest_->is_dim(name);
-  bool is_attr = array_schema_latest_->is_attr(name);
+  bool is_dim = array_schema_latest().is_dim(name);
+  bool is_attr = array_schema_latest().is_attr(name);
 
   // Check if attribute/dimension exists
   if (name != constants::coords && !is_dim && !is_attr) {
@@ -731,7 +707,7 @@ Status Array::get_max_buffer_size(
   }
 
   // Check if attribute/dimension is fixed sized
-  if (array_schema_latest_->var_size(name)) {
+  if (array_schema_latest().var_size(name)) {
     return LOG_STATUS(Status_ArrayError(
         std::string("Cannot get max buffer size; Attribute/Dimension '") +
         name + "' is var-sized"));
@@ -772,14 +748,14 @@ Status Array::get_max_buffer_size(
   }
 
   // Not applicable to heterogeneous domains
-  if (!array_schema_latest_->domain().all_dims_same_type()) {
+  if (!array_schema_latest().domain().all_dims_same_type()) {
     return LOG_STATUS(
         Status_ArrayError("Cannot get max buffer size; Function not "
                           "applicable to heterogeneous domains"));
   }
 
   // Not applicable to variable-sized dimensions
-  if (!array_schema_latest_->domain().all_dims_fixed()) {
+  if (!array_schema_latest().domain().all_dims_fixed()) {
     return LOG_STATUS(Status_ArrayError(
         "Cannot get max buffer size; Function not "
         "applicable to domains with variable-sized dimensions"));
@@ -796,7 +772,7 @@ Status Array::get_max_buffer_size(
   }
 
   // Check if attribute/dimension is var-sized
-  if (!array_schema_latest_->var_size(name)) {
+  if (!array_schema_latest().var_size(name)) {
     return LOG_STATUS(Status_ArrayError(
         std::string("Cannot get max buffer size; Attribute/Dimension '") +
         name + "' is fixed-sized"));
@@ -809,10 +785,6 @@ Status Array::get_max_buffer_size(
   return Status::Ok();
 }
 
-const EncryptionKey& Array::get_encryption_key() const {
-  return *encryption_key_;
-}
-
 Status Array::reopen() {
   // Note: Array will only reopen for reads. This is why we are checking the
   // timestamp for the array directory and not new components. This needs to be
@@ -857,13 +829,18 @@ Status Array::reopen(uint64_t timestamp_start, uint64_t timestamp_end) {
   clear_last_max_buffer_sizes();
 
   // Reopen metadata.
-  fragment_metadata_.clear();
-  metadata_.clear();
-  metadata_loaded_ = false;
-
-  // Reset the non-empty domain - may be different.
-  non_empty_domain_.clear();
-  non_empty_domain_computed_ = false;
+  auto key = opened_array_->encryption_key();
+  opened_array_ = make_shared<OpenedArray>(
+      HERE(),
+      resources_,
+      memory_tracker_,
+      array_uri_,
+      key->encryption_type(),
+      key->key().data(),
+      static_cast<uint32_t>(key->key().size()),
+      this->timestamp_start(),
+      timestamp_end_opened_at(),
+      is_remote());
 
   // Use open to reopen a remote array.
   if (remote_) {
@@ -877,30 +854,33 @@ Status Array::reopen(uint64_t timestamp_start, uint64_t timestamp_end) {
 
     return open(
         query_type_,
-        encryption_key_->encryption_type(),
-        encryption_key_->key().data(),
-        encryption_key_->key().size());
+        encryption_key()->encryption_type(),
+        encryption_key()->key().data(),
+        encryption_key()->key().size());
   }
 
   // Reload the array directory in READ mode (reopen only supports reads).
   try {
     {
       auto timer_se = resources_.stats().start_timer("array_reopen_directory");
-      array_dir_ = ArrayDirectory(
+      set_array_directory(ArrayDirectory(
           resources_,
           array_uri_,
           array_dir_timestamp_start_,
           array_dir_timestamp_end_,
           query_type_ == QueryType::READ ? ArrayDirectoryMode::READ :
-                                           ArrayDirectoryMode::SCHEMA_ONLY);
+                                           ArrayDirectoryMode::SCHEMA_ONLY));
     }
   } catch (const std::logic_error& le) {
     return LOG_STATUS(Status_ArrayDirectoryError(le.what()));
   }
 
   // Reopen the array and update private variables.
-  std::tie(array_schema_latest_, array_schemas_all_, fragment_metadata_) =
+  auto&& [array_schema_latest, array_schemas_all, fragment_metadata] =
       open_for_reads();
+  set_array_schema_latest(array_schema_latest);
+  set_array_schemas_all(std::move(array_schemas_all));
+  set_fragment_metadata(std::move(fragment_metadata));
 
   return Status::Ok();
 }
@@ -931,7 +911,7 @@ void Array::delete_metadata(const char* key) {
     throw ArrayException("Cannot delete metadata. Key cannot be null");
   }
 
-  metadata_.del(key);
+  opened_array_->metadata().del(key);
 }
 
 void Array::put_metadata(
@@ -962,7 +942,7 @@ void Array::put_metadata(
     throw ArrayException("Cannot put metadata; Value type cannot be ANY");
   }
 
-  metadata_.put(key, value_type, value_num, value);
+  opened_array_->metadata().put(key, value_type, value_num, value);
 }
 
 void Array::get_metadata(
@@ -987,11 +967,11 @@ void Array::get_metadata(
   }
 
   // Load array metadata, if not loaded yet
-  if (!metadata_loaded_) {
+  if (!metadata_loaded()) {
     throw_if_not_ok(load_metadata());
   }
 
-  metadata_.get(key, value_type, value_num, value);
+  opened_array_->metadata().get(key, value_type, value_num, value);
 }
 
 void Array::get_metadata(
@@ -1013,11 +993,12 @@ void Array::get_metadata(
   }
 
   // Load array metadata, if not loaded yet
-  if (!metadata_loaded_) {
+  if (!metadata_loaded()) {
     throw_if_not_ok(load_metadata());
   }
 
-  metadata_.get(index, key, key_len, value_type, value_num, value);
+  opened_array_->metadata().get(
+      index, key, key_len, value_type, value_num, value);
 }
 
 uint64_t Array::metadata_num() {
@@ -1033,11 +1014,11 @@ uint64_t Array::metadata_num() {
   }
 
   // Load array metadata, if not loaded yet
-  if (!metadata_loaded_) {
+  if (!metadata_loaded()) {
     throw_if_not_ok(load_metadata());
   }
 
-  return metadata_.num();
+  return opened_array_->metadata().num();
 }
 
 std::optional<Datatype> Array::metadata_type(const char* key) {
@@ -1058,47 +1039,33 @@ std::optional<Datatype> Array::metadata_type(const char* key) {
   }
 
   // Load array metadata, if not loaded yet
-  if (!metadata_loaded_) {
+  if (!metadata_loaded()) {
     throw_if_not_ok(load_metadata());
   }
 
-  return metadata_.metadata_type(key);
+  return opened_array_->metadata().metadata_type(key);
 }
 
 Metadata* Array::unsafe_metadata() {
-  return &metadata_;
+  return &opened_array_->metadata();
 }
 
-Status Array::metadata(Metadata** metadata) {
+Metadata& Array::metadata() {
   // Load array metadata for array opened for reads, if not loaded yet
-  if (query_type_ == QueryType::READ && !metadata_loaded_) {
-    RETURN_NOT_OK(load_metadata());
+  if (query_type_ == QueryType::READ && !metadata_loaded()) {
+    throw_if_not_ok(load_metadata());
   }
 
-  *metadata = &metadata_;
-
-  return Status::Ok();
+  return opened_array_->metadata();
 }
 
-NDRange* Array::loaded_non_empty_domain() {
-  return &non_empty_domain_;
-}
-
-tuple<Status, optional<const NDRange>> Array::non_empty_domain() {
-  if (!non_empty_domain_computed_) {
+const NDRange Array::non_empty_domain() {
+  if (!non_empty_domain_computed()) {
     // Compute non-empty domain
-    RETURN_NOT_OK_TUPLE(compute_non_empty_domain(), nullopt);
+    throw_if_not_ok(compute_non_empty_domain());
   }
 
-  return {Status::Ok(), non_empty_domain_};
-}
-
-void Array::set_non_empty_domain(const NDRange& non_empty_domain) {
-  non_empty_domain_ = non_empty_domain;
-}
-
-MemoryTracker* Array::memory_tracker() {
-  return &memory_tracker_;
+  return loaded_non_empty_domain();
 }
 
 bool Array::serialize_non_empty_domain() const {
@@ -1194,12 +1161,17 @@ std::unordered_map<std::string, uint64_t> Array::get_average_var_cell_sizes()
     const {
   std::unordered_map<std::string, uint64_t> ret;
 
+  // Keep the current opened array alive for the duration of this call.
+  auto opened_array = opened_array_;
+  auto& fragment_metadata = opened_array->fragment_metadata();
+  auto& array_schema_latest = opened_array->array_schema_latest();
+
   // Find the names of the var-sized dimensions or attributes.
   std::vector<std::string> var_names;
 
   // Start with dimensions.
-  for (unsigned d = 0; d < array_schema_latest_->dim_num(); d++) {
-    auto dim = array_schema_latest_->dimension_ptr(d);
+  for (unsigned d = 0; d < array_schema_latest.dim_num(); d++) {
+    auto dim = array_schema_latest.dimension_ptr(d);
     if (dim->var_size()) {
       var_names.emplace_back(dim->name());
       ret.emplace(dim->name(), 0);
@@ -1207,7 +1179,7 @@ std::unordered_map<std::string, uint64_t> Array::get_average_var_cell_sizes()
   }
 
   // Now attributes.
-  for (auto& attr : array_schema_latest_->attributes()) {
+  for (auto& attr : array_schema_latest.attributes()) {
     if (attr->var_size()) {
       var_names.emplace_back(attr->name());
       ret.emplace(attr->name(), 0);
@@ -1219,17 +1191,17 @@ std::unordered_map<std::string, uint64_t> Array::get_average_var_cell_sizes()
     throw_if_not_ok(parallel_for(
         &resources_.compute_tp(),
         0,
-        fragment_metadata_.size(),
+        fragment_metadata.size(),
         [&](const unsigned f) {
           // Gracefully skip loading tile sizes for attributes added in schema
           // evolution that do not exists in this fragment.
-          const auto& schema = fragment_metadata_[f]->array_schema();
+          const auto& schema = fragment_metadata[f]->array_schema();
           if (!schema->is_field(var_name)) {
             return Status::Ok();
           }
 
-          fragment_metadata_[f]->load_tile_var_sizes(
-              *encryption_key_, var_name);
+          fragment_metadata[f]->load_tile_var_sizes(
+              *encryption_key(), var_name);
           return Status::Ok();
         }));
   }
@@ -1242,18 +1214,18 @@ std::unordered_map<std::string, uint64_t> Array::get_average_var_cell_sizes()
         auto& var_name = var_names[n];
 
         // Sum the total tile size and cell num for each fragments.
-        for (unsigned f = 0; f < fragment_metadata_.size(); f++) {
+        for (unsigned f = 0; f < fragment_metadata.size(); f++) {
           // Skip computation for fields that don't exist for a particular
           // fragment.
-          const auto& schema = fragment_metadata_[f]->array_schema();
+          const auto& schema = fragment_metadata[f]->array_schema();
           if (!schema->is_field(var_name)) {
             continue;
           }
 
           // Go through all tiles.
-          for (uint64_t t = 0; t < fragment_metadata_[f]->tile_num(); t++) {
-            total_size += fragment_metadata_[f]->tile_var_size(var_name, t);
-            cell_num += fragment_metadata_[f]->cell_num(t);
+          for (uint64_t t = 0; t < fragment_metadata[f]->tile_num(); t++) {
+            total_size += fragment_metadata[f]->tile_var_size(var_name, t);
+            cell_num += fragment_metadata[f]->cell_num(t);
           }
         }
 
@@ -1294,7 +1266,8 @@ Array::open_for_reads_without_fragments() {
       "array_open_read_without_fragments_load_schemas");
 
   // Load array schemas
-  auto result = array_dir_.load_array_schemas(*encryption_key());
+  auto result =
+      array_directory().load_array_schemas(*encryption_key(), memory_tracker_);
 
   auto version = std::get<0>(result)->version();
   ensure_supported_schema_version_for_read(version);
@@ -1319,7 +1292,7 @@ Array::open_for_writes() {
 
   // Load array schemas
   auto&& [array_schema_latest, array_schemas_all] =
-      array_dir_.load_array_schemas(*encryption_key_);
+      array_directory().load_array_schemas(*encryption_key(), memory_tracker_);
 
   // If building experimentally, this library should not be able to
   // write to newer-versioned or older-versioned arrays
@@ -1363,16 +1336,16 @@ void Array::clear_last_max_buffer_sizes() {
 
 Status Array::compute_max_buffer_sizes(const void* subarray) {
   // Applicable only to domains where all dimensions have the same type
-  if (!array_schema_latest_->domain().all_dims_same_type()) {
+  if (!array_schema_latest().domain().all_dims_same_type()) {
     return LOG_STATUS(
         Status_ArrayError("Cannot compute max buffer sizes; Inapplicable when "
                           "dimension domains have different types"));
   }
 
   // Allocate space for max buffer sizes subarray
-  auto dim_num = array_schema_latest_->dim_num();
+  auto dim_num = array_schema_latest().dim_num();
   auto coord_size{
-      array_schema_latest_->domain().dimension_ptr(0)->coord_size()};
+      array_schema_latest().domain().dimension_ptr(0)->coord_size()};
   auto subarray_size = 2 * dim_num * coord_size;
   last_max_buffer_sizes_subarray_.resize(subarray_size);
 
@@ -1383,7 +1356,7 @@ Status Array::compute_max_buffer_sizes(const void* subarray) {
     last_max_buffer_sizes_.clear();
 
     // Get all attributes and coordinates
-    auto attributes = array_schema_latest_->attributes();
+    auto& attributes = array_schema_latest().attributes();
     last_max_buffer_sizes_.clear();
     for (const auto& attr : attributes)
       last_max_buffer_sizes_[attr->name()] =
@@ -1392,7 +1365,7 @@ Status Array::compute_max_buffer_sizes(const void* subarray) {
         std::pair<uint64_t, uint64_t>(0, 0);
     for (unsigned d = 0; d < dim_num; ++d)
       last_max_buffer_sizes_
-          [array_schema_latest_->domain().dimension_ptr(d)->name()] =
+          [array_schema_latest().domain().dimension_ptr(d)->name()] =
               std::pair<uint64_t, uint64_t>(0, 0);
 
     RETURN_NOT_OK(compute_max_buffer_sizes(subarray, &last_max_buffer_sizes_));
@@ -1416,64 +1389,68 @@ Status Array::compute_max_buffer_sizes(
     }
 
     return rest_client->get_array_max_buffer_sizes(
-        array_uri_, *(array_schema_latest_.get()), subarray, buffer_sizes);
+        array_uri_, array_schema_latest(), subarray, buffer_sizes);
   }
 
+  // Keep the current opened array alive for the duration of this call.
+  auto opened_array = opened_array_;
+  auto& fragment_metadata = opened_array->fragment_metadata();
+  auto& array_schema_latest = opened_array->array_schema_latest();
+
   // Return if there are no metadata
-  if (fragment_metadata_.empty()) {
+  if (fragment_metadata.empty()) {
     return Status::Ok();
   }
 
   // First we calculate a rough upper bound. Especially for dense
   // arrays, this will not be accurate, as it accounts only for the
   // non-empty regions of the subarray.
-  for (auto& meta : fragment_metadata_) {
-    meta->add_max_buffer_sizes(*encryption_key_, subarray, buffer_sizes);
+  for (auto& meta : fragment_metadata) {
+    meta->add_max_buffer_sizes(*encryption_key(), subarray, buffer_sizes);
   }
 
   // Prepare an NDRange for the subarray
-  auto dim_num = array_schema_latest_->dim_num();
+  auto dim_num = array_schema_latest.dim_num();
   NDRange sub(dim_num);
   auto sub_ptr = (const unsigned char*)subarray;
   uint64_t offset = 0;
   for (unsigned d = 0; d < dim_num; ++d) {
-    auto r_size{2 * array_schema_latest_->dimension_ptr(d)->coord_size()};
+    auto r_size{2 * array_schema_latest.dimension_ptr(d)->coord_size()};
     sub[d] = Range(&sub_ptr[offset], r_size);
     offset += r_size;
   }
 
   // Rectify bound for dense arrays
-  if (array_schema_latest_->dense()) {
-    auto cell_num = array_schema_latest_->domain().cell_num(sub);
+  if (array_schema_latest.dense()) {
+    auto cell_num = array_schema_latest.domain().cell_num(sub);
     // `cell_num` becomes 0 when `subarray` is huge, leading to a
     // `uint64_t` overflow.
     if (cell_num != 0) {
       for (auto& it : *buffer_sizes) {
-        if (array_schema_latest_->var_size(it.first)) {
+        if (array_schema_latest.var_size(it.first)) {
           it.second.first = cell_num * constants::cell_var_offset_size;
           it.second.second +=
-              cell_num * datatype_size(array_schema_latest_->type(it.first));
+              cell_num * datatype_size(array_schema_latest.type(it.first));
         } else {
-          it.second.first =
-              cell_num * array_schema_latest_->cell_size(it.first);
+          it.second.first = cell_num * array_schema_latest.cell_size(it.first);
         }
       }
     }
   }
 
   // Rectify bound for sparse arrays with integer domain, without duplicates
-  if (!array_schema_latest_->dense() && !array_schema_latest_->allows_dups() &&
-      array_schema_latest_->domain().all_dims_int()) {
-    auto cell_num = array_schema_latest_->domain().cell_num(sub);
+  if (!array_schema_latest.dense() && !array_schema_latest.allows_dups() &&
+      array_schema_latest.domain().all_dims_int()) {
+    auto cell_num = array_schema_latest.domain().cell_num(sub);
     // `cell_num` becomes 0 when `subarray` is huge, leading to a
     // `uint64_t` overflow.
     if (cell_num != 0) {
       for (auto& it : *buffer_sizes) {
-        if (!array_schema_latest_->var_size(it.first)) {
+        if (!array_schema_latest.var_size(it.first)) {
           // Check for overflow
           uint64_t new_size =
-              cell_num * array_schema_latest_->cell_size(it.first);
-          if (new_size / array_schema_latest_->cell_size((it.first)) !=
+              cell_num * array_schema_latest.cell_size(it.first);
+          if (new_size / array_schema_latest.cell_size((it.first)) !=
               cell_num) {
             continue;
           }
@@ -1489,14 +1466,14 @@ Status Array::compute_max_buffer_sizes(
 }
 
 void Array::do_load_metadata() {
-  if (!array_dir_.loaded()) {
+  if (!array_directory().loaded()) {
     throw ArrayException(
         "Cannot load metadata; array directory is not loaded.");
   }
   auto timer_se = resources_.stats().start_timer("sm_load_array_metadata");
 
   // Determine which array metadata to load
-  const auto& array_metadata_to_load = array_dir_.array_meta_uris();
+  const auto& array_metadata_to_load = array_directory().array_meta_uris();
 
   auto metadata_num = array_metadata_to_load.size();
   std::vector<shared_ptr<Tile>> metadata_tiles(metadata_num);
@@ -1504,8 +1481,8 @@ void Array::do_load_metadata() {
       parallel_for(&resources_.compute_tp(), 0, metadata_num, [&](size_t m) {
         const auto& uri = array_metadata_to_load[m].uri_;
 
-        auto&& tile = GenericTileIO::load(resources_, uri, 0, *encryption_key_);
-        metadata_tiles[m] = tdb::make_shared<Tile>(HERE(), std::move(tile));
+        metadata_tiles[m] = GenericTileIO::load(
+            resources_, uri, 0, *encryption_key(), memory_tracker_);
 
         return Status::Ok();
       }));
@@ -1517,10 +1494,10 @@ void Array::do_load_metadata() {
   }
   resources_.stats().add_counter("read_array_meta_size", meta_size);
 
-  metadata_ = Metadata::deserialize(metadata_tiles);
+  opened_array_->metadata() = Metadata::deserialize(metadata_tiles);
 
   // Sets the loaded metadata URIs
-  metadata_.set_loaded_metadata_uris(array_metadata_to_load);
+  opened_array_->metadata().set_loaded_metadata_uris(array_metadata_to_load);
 }
 
 Status Array::load_metadata() {
@@ -1538,7 +1515,7 @@ Status Array::load_metadata() {
   } else {
     do_load_metadata();
   }
-  metadata_loaded_ = true;
+  set_metadata_loaded(true);
   return Status::Ok();
 }
 
@@ -1551,14 +1528,14 @@ Status Array::load_remote_non_empty_domain() {
     }
     RETURN_NOT_OK(rest_client->get_array_non_empty_domain(
         this, array_dir_timestamp_start_, timestamp_end_opened_at()));
-    non_empty_domain_computed_ = true;
+    set_non_empty_domain_computed(true);
   }
   return Status::Ok();
 }
 
-ArrayDirectory& Array::load_array_directory() {
-  if (array_dir_.loaded()) {
-    return array_dir_;
+const ArrayDirectory& Array::load_array_directory() {
+  if (array_directory().loaded()) {
+    return array_directory();
   }
 
   if (remote_) {
@@ -1571,45 +1548,51 @@ ArrayDirectory& Array::load_array_directory() {
                   ArrayDirectoryMode::SCHEMA_ONLY :
                   ArrayDirectoryMode::READ;
 
-  array_dir_ = ArrayDirectory(
+  set_array_directory(ArrayDirectory(
       resources_,
       array_uri_,
       array_dir_timestamp_start_,
       array_dir_timestamp_end_,
-      mode);
+      mode));
 
-  return array_dir_;
+  return array_directory();
 }
 
 Status Array::compute_non_empty_domain() {
+  // Keep the current opened array alive for the duration of this call.
+  auto opened_array = opened_array_;
+  auto& fragment_metadata = opened_array->fragment_metadata();
+  auto& array_schema_latest = opened_array->array_schema_latest();
+  auto& non_empty_domain = opened_array->non_empty_domain();
+
   if (remote_) {
     RETURN_NOT_OK(load_remote_non_empty_domain());
-  } else if (!fragment_metadata_.empty()) {
-    const auto& frag0_dom = fragment_metadata_[0]->non_empty_domain();
-    non_empty_domain_.assign(frag0_dom.begin(), frag0_dom.end());
+  } else if (!fragment_metadata.empty()) {
+    const auto& frag0_dom = fragment_metadata[0]->non_empty_domain();
+    non_empty_domain.assign(frag0_dom.begin(), frag0_dom.end());
 
-    auto metadata_num = fragment_metadata_.size();
+    auto metadata_num = fragment_metadata.size();
     for (size_t j = 1; j < metadata_num; ++j) {
-      const auto& meta_dom = fragment_metadata_[j]->non_empty_domain();
+      const auto& meta_dom = fragment_metadata[j]->non_empty_domain();
       // Validate that this fragment's non-empty domain is set
       // It should _always_ be set, however we've seen cases where disk
       // corruption or other out-of-band activity can cause the fragment to be
       // corrupt for these cases we want to check to prevent any segfaults
       // later.
       if (!meta_dom.empty()) {
-        array_schema_latest_->domain().expand_ndrange(
-            meta_dom, &non_empty_domain_);
+        array_schema_latest.domain().expand_ndrange(
+            meta_dom, &non_empty_domain);
       } else {
         // If the fragment's non-empty domain is indeed empty, lets log it so
         // the user gets a message warning that this fragment might be corrupt
         // Note: LOG_STATUS only prints if TileDB is built in verbose mode.
         LOG_STATUS_NO_RETURN_VALUE(Status_ArrayError(
             "Non empty domain unexpectedly empty for fragment: " +
-            fragment_metadata_[j]->fragment_uri().to_string()));
+            fragment_metadata[j]->fragment_uri().to_string()));
       }
     }
   }
-  non_empty_domain_computed_ = true;
+  opened_array->non_empty_domain_computed() = true;
   return Status::Ok();
 }
 
@@ -1631,6 +1614,31 @@ void Array::set_array_open(const QueryType& query_type) {
   is_open_ = true;
 }
 
+void Array::set_serialized_array_open() {
+  is_open_ = true;
+
+  opened_array_ = make_shared<OpenedArray>(
+      HERE(),
+      resources_,
+      memory_tracker_,
+      array_uri_,
+      EncryptionType::NO_ENCRYPTION,
+      nullptr,
+      0,
+      timestamp_start(),
+      timestamp_end_opened_at(),
+      array_uri_.is_tiledb());
+}
+
+void Array::set_query_type(QueryType query_type) {
+  query_type_ = query_type;
+  if (query_type_ == QueryType::READ) {
+    memory_tracker_->set_type(MemoryTrackerType::ARRAY_READ);
+  } else {
+    memory_tracker_->set_type(MemoryTrackerType::ARRAY_WRITE);
+  }
+}
+
 void Array::set_array_closed() {
   std::lock_guard<std::mutex> lock(mtx_);
 
@@ -1684,5 +1692,4 @@ void ensure_supported_schema_version_for_read(format_version_t version) {
   }
 }
 
-}  // namespace sm
-}  // namespace tiledb
+}  // namespace tiledb::sm
