@@ -35,6 +35,7 @@
 #include "tiledb/common/heap_memory.h"
 #include "tiledb/common/logger.h"
 #include "tiledb/common/memory.h"
+#include "tiledb/common/memory_tracker.h"
 #include "tiledb/sm/array/array.h"
 #include "tiledb/sm/array_schema/dimension_label.h"
 #include "tiledb/sm/enums/query_status.h"
@@ -66,14 +67,6 @@ using namespace tiledb::sm::stats;
 namespace tiledb {
 namespace sm {
 
-/** Class for query status exceptions. */
-class QueryStatusException : public StatusException {
- public:
-  explicit QueryStatusException(const std::string& msg)
-      : StatusException("Query", msg) {
-  }
-};
-
 /* ****************************** */
 /*   CONSTRUCTORS & DESTRUCTORS   */
 /* ****************************** */
@@ -82,7 +75,9 @@ Query::Query(
     StorageManager* storage_manager,
     shared_ptr<Array> array,
     optional<std::string> fragment_name)
-    : array_shared_(array)
+    : query_memory_tracker_(
+          storage_manager->resources().create_memory_tracker())
+    , array_shared_(array)
     , array_(array_shared_.get())
     , opened_array_(array->opened_array())
     , array_schema_(array->array_schema_latest_ptr())
@@ -110,9 +105,16 @@ Query::Query(
     , is_dimension_label_ordered_read_(false)
     , dimension_label_increasing_(true)
     , fragment_size_(std::numeric_limits<uint64_t>::max())
-    , query_remote_buffer_storage_(std::nullopt) {
+    , query_remote_buffer_storage_(std::nullopt)
+    , default_channel_{make_shared<QueryChannel>(HERE(), *this, 0)} {
   assert(array->is_open());
 
+  if (array->get_query_type() == QueryType::READ) {
+    query_memory_tracker_->set_type(MemoryTrackerType::QUERY_READ);
+  } else {
+    query_memory_tracker_->set_type(MemoryTrackerType::QUERY_WRITE);
+  }
+
   subarray_ = Subarray(array_, layout_, stats_, logger_);
 
   fragment_metadata_ = array->fragment_metadata();
@@ -827,6 +829,7 @@ Status Query::process() {
 
     if (!only_dim_label_query()) {
       if (strategy_ != nullptr) {
+        // The strategy destructor should reset its own Stats object here
         dynamic_cast<StrategyBase*>(strategy_.get())->stats()->reset();
         strategy_ = nullptr;
       }
@@ -917,6 +920,7 @@ Status Query::reset_strategy_with_layout(
     Layout layout, bool force_legacy_reader) {
   force_legacy_reader_ = force_legacy_reader;
   if (strategy_ != nullptr) {
+    // The strategy destructor should reset its own Stats object here
     dynamic_cast<StrategyBase*>(strategy_.get())->stats()->reset();
     strategy_ = nullptr;
   }
@@ -1663,6 +1667,10 @@ stats::Stats* Query::stats() const {
   return stats_;
 }
 
+void Query::set_stats(const stats::StatsData& data) {
+  stats_->populate_with_data(data);
+}
+
 shared_ptr<Buffer> Query::rest_scratch() const {
   return rest_scratch_;
 }
@@ -1791,6 +1799,8 @@ bool Query::is_aggregate(std::string output_field_name) const {
 
 Status Query::create_strategy(bool skip_checks_serialization) {
   auto params = StrategyParams(
+      array_->memory_tracker(),
+      query_memory_tracker_,
       storage_manager_,
       opened_array_,
       config_,
@@ -1800,8 +1810,7 @@ Status Query::create_strategy(bool skip_checks_serialization) {
       layout_,
       condition_,
       default_channel_aggregates_,
-      skip_checks_serialization,
-      array_->memory_tracker());
+      skip_checks_serialization);
   if (type_ == QueryType::WRITE || type_ == QueryType::MODIFY_EXCLUSIVE) {
     if (layout_ == Layout::COL_MAJOR || layout_ == Layout::ROW_MAJOR) {
       if (!array_schema_->dense()) {
@@ -2138,5 +2147,9 @@ void Query::copy_aggregates_data_to_user_buffer() {
   }
 }
 
+RestClient* Query::rest_client() const {
+  return storage_manager_->rest_client();
+}
+
 }  // namespace sm
 }  // namespace tiledb
