@@ -205,6 +205,60 @@ class Subarray {
     uint64_t size_validity_;
   };
 
+  /**
+   * Wrapper for optional<tuple<std::string, RangeSetAndSuperset>> for
+   * cleaner data access.
+   */
+  struct LabelRangeSubset {
+   public:
+    /**
+     * Default constructor is not C.41.
+     **/
+    LabelRangeSubset() = delete;
+
+    /**
+     * Constructor
+     *
+     * @param ref Dimension label the ranges will be set on.
+     * @param coalesce_ranges Set if ranges should be combined when adjacent.
+     */
+    LabelRangeSubset(const DimensionLabel& ref, bool coalesce_ranges = true);
+
+    /**
+     * Constructor
+     *
+     * @param name The name of the dimension label the ranges will be set on.
+     * @param type The type of the label the ranges will be set on.
+     * @param coalesce_ranges Set if ranges should be combined when adjacent.
+     */
+    LabelRangeSubset(
+        const std::string& name, Datatype type, bool coalesce_ranges = true);
+
+    /**
+     * Constructor
+     *
+     * @param name The name of the dimension label the ranges will be set on.
+     * @param type The type of the label the ranges will be set on.
+     * @param ranges The range subset for the dimension label.
+     * @param coalesce_ranges Set if ranges should be combined when adjacent.
+     */
+    LabelRangeSubset(
+        const std::string& name,
+        Datatype type,
+        std::vector<Range> ranges,
+        bool coalesce_ranges = true);
+
+    inline const std::vector<Range>& get_ranges() const {
+      return ranges_.ranges();
+    }
+
+    /** Name of the dimension label. */
+    std::string name_;
+
+    /** The ranges set on the dimension label. */
+    RangeSetAndSuperset ranges_;
+  };
+
   /* ********************************* */
   /*     CONSTRUCTORS & DESTRUCTORS    */
   /* ********************************* */
@@ -707,7 +761,7 @@ class Subarray {
    */
   Status precompute_all_ranges_tile_overlap(
       ThreadPool* const compute_tp,
-      std::vector<FragIdx>& frag_tile_idx,
+      const std::vector<FragIdx>& frag_tile_idx,
       ITileRange* tile_ranges);
 
   /**
@@ -1276,7 +1330,15 @@ class Subarray {
       std::vector<uint64_t>* end_coords) const;
 
   /** Returns `stats_`. */
-  stats::Stats* stats() const;
+  const stats::Stats& stats() const;
+
+  /**
+   * Populate the owned stats instance with data.
+   * To be removed when the class will get a C41 constructor.
+   *
+   * @param data Data to populate the stats with.
+   */
+  void set_stats(const stats::StatsData& data);
 
   /** Stores a vector of 1D ranges per dimension. */
   std::vector<std::vector<uint64_t>> original_range_idx_;
@@ -1306,6 +1368,20 @@ class Subarray {
    */
   void add_default_label_ranges(dimension_size_type dim_num);
 
+  /**
+   * Reset ranges to default if possible before a read operation for sparse
+   * reads. We have a lot of optimizations in the sparse readers when no ranges
+   * are specified. Python will set ranges that are equal to the non empty
+   * domain, which will negate those optimizations. When the non empty domain is
+   * computed for the array, it is low performance cost to see if the ranges set
+   * are equal to the non empty domain. If they are, we can reset them to be
+   * default.
+   */
+  void reset_default_ranges();
+
+  /** Loads the R-Trees of all relevant fragments in parallel. */
+  Status load_relevant_fragment_rtrees(ThreadPool* compute_tp) const;
+
  private:
   /* ********************************* */
   /*        PRIVATE DATA TYPES         */
@@ -1332,46 +1408,6 @@ class Subarray {
     uint64_t range_len_;
   };
 
-  /**
-   * Wrapper for optional<tuple<std::string, RangeSetAndSuperset>> for
-   * cleaner data access.
-   */
-  struct LabelRangeSubset {
-   public:
-    /**
-     * Default constructor is not C.41.
-     **/
-    LabelRangeSubset() = delete;
-
-    /**
-     * Constructor
-     *
-     * @param ref Dimension label the ranges will be set on.
-     * @param coalesce_ranges Set if ranges should be combined when adjacent.
-     */
-    LabelRangeSubset(const DimensionLabel& ref, bool coalesce_ranges = true);
-
-    /**
-     * Constructor
-     *
-     * @param name The name of the dimension label the ranges will be set on.
-     * @param type The type of the label the ranges will be set on.
-     * @param coalesce_ranges Set if ranges should be combined when adjacent.
-     */
-    LabelRangeSubset(
-        const std::string& name, Datatype type, bool coalesce_ranges = true);
-
-    inline const std::vector<Range>& get_ranges() const {
-      return ranges_.ranges();
-    }
-
-    /** Name of the dimension label. */
-    std::string name_;
-
-    /** The ranges set on the dimension label. */
-    RangeSetAndSuperset ranges_;
-  };
-
   /**
    * A hash function capable of hashing std::vector<uint8_t> for use by
    * the tile_coords_map_ unordered_map for caching coords indices.
@@ -1577,9 +1613,6 @@ class Subarray {
    */
   void swap(Subarray& subarray);
 
-  /** Loads the R-Trees of all relevant fragments in parallel. */
-  Status load_relevant_fragment_rtrees(ThreadPool* compute_tp) const;
-
   /**
    * Computes the tile overlap for each range and relevant fragment.
    *
