@@ -31,22 +31,21 @@
  */
 
 #include "tiledb/sm/tile/tile.h"
+#include "tiledb/common/exception/exception.h"
 #include "tiledb/common/heap_memory.h"
-#include "tiledb/common/logger.h"
+#include "tiledb/common/memory_tracker.h"
 #include "tiledb/sm/enums/datatype.h"
 #include "tiledb/storage_format/serialization/serializers.h"
 
-#include <iostream>
-
 using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
 
 /** Class for locally generated status exceptions. */
-class TileStatusException : public StatusException {
+class TileException : public StatusException {
  public:
-  explicit TileStatusException(const std::string& msg)
+  explicit TileException(const std::string& msg)
       : StatusException("Tile", msg) {
   }
 };
@@ -59,23 +58,29 @@ uint64_t WriterTile::max_tile_chunk_size_ = constants::max_tile_chunk_size;
 /*           STATIC API           */
 /* ****************************** */
 
-Tile Tile::from_generic(storage_size_t tile_size) {
-  return {
+shared_ptr<Tile> Tile::from_generic(
+    storage_size_t tile_size, shared_ptr<MemoryTracker> memory_tracker) {
+  return make_shared<Tile>(
+      HERE(),
       0,
       constants::generic_tile_datatype,
       constants::generic_tile_cell_size,
       0,
       tile_size,
       nullptr,
-      0};
+      0,
+      memory_tracker->get_resource(MemoryType::GENERIC_TILE_IO));
 }
 
-WriterTile WriterTile::from_generic(storage_size_t tile_size) {
-  return {
+shared_ptr<WriterTile> WriterTile::from_generic(
+    storage_size_t tile_size, shared_ptr<MemoryTracker> memory_tracker) {
+  return make_shared<WriterTile>(
+      HERE(),
       0,
       constants::generic_tile_datatype,
       constants::generic_tile_cell_size,
-      tile_size};
+      tile_size,
+      memory_tracker->get_resource(MemoryType::GENERIC_TILE_IO));
 }
 
 uint32_t WriterTile::compute_chunk_size(
@@ -87,7 +92,7 @@ uint32_t WriterTile::compute_chunk_size(
   chunk_size64 = chunk_size64 / dim_cell_size * dim_cell_size;
   chunk_size64 = std::max(chunk_size64, dim_cell_size);
   if (chunk_size64 > std::numeric_limits<uint32_t>::max()) {
-    throw TileStatusException("Chunk size exceeds uint32_t");
+    throw TileException("Chunk size exceeds uint32_t");
   }
 
   return static_cast<uint32_t>(chunk_size64);
@@ -105,8 +110,10 @@ TileBase::TileBase(
     const format_version_t format_version,
     const Datatype type,
     const uint64_t cell_size,
-    const uint64_t size)
-    : data_(static_cast<char*>(tdb_malloc(size)), tiledb_free)
+    const uint64_t size,
+    tdb::pmr::memory_resource* resource)
+    : resource_(resource)
+    , data_(tdb::pmr::make_unique<std::byte>(resource_, size))
     , size_(size)
     , cell_size_(cell_size)
     , format_version_(format_version)
@@ -121,19 +128,24 @@ TileBase::TileBase(
   }
 }
 
-TileBase::TileBase(TileBase&& tile)
-    : data_(std::move(tile.data_))
-    , size_(std::move(tile.size_))
-    , cell_size_(std::move(tile.cell_size_))
-    , format_version_(std::move(tile.format_version_))
-    , type_(std::move(tile.type_)) {
-}
-
-TileBase& TileBase::operator=(TileBase&& tile) {
-  // Swap with the argument
-  swap(tile);
-
-  return *this;
+Tile::Tile(
+    const format_version_t format_version,
+    const Datatype type,
+    const uint64_t cell_size,
+    const unsigned int zipped_coords_dim_num,
+    const uint64_t size,
+    void* filtered_data,
+    uint64_t filtered_size,
+    shared_ptr<MemoryTracker> memory_tracker)
+    : Tile(
+          format_version,
+          type,
+          cell_size,
+          zipped_coords_dim_num,
+          size,
+          filtered_data,
+          filtered_size,
+          memory_tracker->get_resource(MemoryType::TILE_DATA)) {
 }
 
 Tile::Tile(
@@ -143,83 +155,61 @@ Tile::Tile(
     const unsigned int zipped_coords_dim_num,
     const uint64_t size,
     void* filtered_data,
-    uint64_t filtered_size)
-    : TileBase(format_version, type, cell_size, size)
+    uint64_t filtered_size,
+    tdb::pmr::memory_resource* resource)
+    : TileBase(format_version, type, cell_size, size, resource)
     , zipped_coords_dim_num_(zipped_coords_dim_num)
     , filtered_data_(filtered_data)
     , filtered_size_(filtered_size) {
 }
 
-Tile::Tile(Tile&& tile)
-    : TileBase(std::move(tile))
-    , zipped_coords_dim_num_(std::move(tile.zipped_coords_dim_num_))
-    , filtered_data_(std::move(tile.filtered_data_))
-    , filtered_size_(std::move(tile.filtered_size_)) {
-}
-
-Tile& Tile::operator=(Tile&& tile) {
-  // Swap with the argument
-  swap(tile);
-
-  return *this;
-}
-
 WriterTile::WriterTile(
     const format_version_t format_version,
     const Datatype type,
     const uint64_t cell_size,
-    const uint64_t size)
-    : TileBase(format_version, type, cell_size, size)
+    const uint64_t size,
+    shared_ptr<MemoryTracker> memory_tracker)
+    : TileBase(
+          format_version,
+          type,
+          cell_size,
+          size,
+          memory_tracker->get_resource(MemoryType::TILE_WRITER_DATA))
     , filtered_buffer_(0) {
 }
 
-WriterTile::WriterTile(WriterTile&& tile)
-    : TileBase(std::move(tile))
-    , filtered_buffer_(std::move(tile.filtered_buffer_)) {
-}
-
-WriterTile& WriterTile::operator=(WriterTile&& tile) {
-  // Swap with the argument
-  swap(tile);
-
-  return *this;
+WriterTile::WriterTile(
+    const format_version_t format_version,
+    const Datatype type,
+    const uint64_t cell_size,
+    const uint64_t size,
+    tdb::pmr::memory_resource* resource)
+    : TileBase(format_version, type, cell_size, size, resource)
+    , filtered_buffer_(0) {
 }
 
 /* ****************************** */
 /*               API              */
 /* ****************************** */
 
-void TileBase::swap(TileBase& tile) {
-  std::swap(size_, tile.size_);
-  std::swap(data_, tile.data_);
-  std::swap(cell_size_, tile.cell_size_);
-  std::swap(format_version_, tile.format_version_);
-  std::swap(type_, tile.type_);
-}
-
-Status TileBase::read(
+void TileBase::read(
     void* const buffer, const uint64_t offset, const uint64_t nbytes) const {
   if (nbytes > size_ - offset) {
-    return LOG_STATUS(Status_TileError(
-        "Read tile overflow; may not read beyond buffer size"));
+    throw TileException("Read tile overflow; may not read beyond buffer size");
   }
   std::memcpy(buffer, data_.get() + offset, nbytes);
-  return Status::Ok();
 }
 
-Status TileBase::write(const void* data, uint64_t offset, uint64_t nbytes) {
+void TileBase::write(const void* data, uint64_t offset, uint64_t nbytes) {
   if (nbytes > size_ - offset) {
-    return LOG_STATUS(
-        Status_TileError("Write tile overflow; would write out of bounds"));
+    throw TileException("Write tile overflow; would write out of bounds");
   }
 
   std::memcpy(data_.get() + offset, data, nbytes);
   size_ = std::max(offset + nbytes, size_);
-
-  return Status::Ok();
 }
 
-Status Tile::zip_coordinates() {
+void Tile::zip_coordinates() {
   assert(zipped_coords_dim_num_ > 0);
 
   // For easy reference
@@ -245,8 +235,6 @@ Status Tile::zip_coordinates() {
 
   // Clean up
   tdb_free((void*)tile_tmp);
-
-  return Status::Ok();
 }
 
 uint64_t Tile::load_chunk_data(ChunkData& chunk_data) {
@@ -256,47 +244,36 @@ uint64_t Tile::load_chunk_data(ChunkData& chunk_data) {
 uint64_t Tile::load_offsets_chunk_data(ChunkData& chunk_data) {
   auto s = size();
   if (s < 8) {
-    throw std::runtime_error("Offsets tile should at least be 8 bytes.");
+    throw TileException("Offsets tile should at least be 8 bytes.");
   }
 
   return load_chunk_data(chunk_data, s - 8);
 }
 
-void Tile::swap(Tile& tile) {
-  TileBase::swap(tile);
-  std::swap(filtered_data_, tile.filtered_data_);
-  std::swap(filtered_size_, tile.filtered_size_);
-  std::swap(zipped_coords_dim_num_, tile.zipped_coords_dim_num_);
-}
-
 void WriterTile::clear_data() {
   data_ = nullptr;
   size_ = 0;
 }
 
-Status WriterTile::write_var(
-    const void* data, uint64_t offset, uint64_t nbytes) {
+void WriterTile::write_var(const void* data, uint64_t offset, uint64_t nbytes) {
   if (size_ - offset < nbytes) {
     auto new_alloc_size = size_ == 0 ? offset + nbytes : size_;
     while (new_alloc_size < offset + nbytes)
       new_alloc_size *= 2;
 
-    auto new_data =
-        static_cast<char*>(tdb_realloc(data_.release(), new_alloc_size));
+    auto new_data = tdb::pmr::make_unique<std::byte>(resource_, new_alloc_size);
+
     if (new_data == nullptr) {
-      return LOG_STATUS(Status_TileError(
-          "Cannot reallocate buffer; Memory allocation failed"));
+      throw TileException("Cannot reallocate buffer; Memory allocation failed");
     }
-    data_.reset(new_data);
+
+    std::memcpy(new_data.get(), data_.get(), std::min(size_, new_alloc_size));
+
+    data_ = std::move(new_data);
     size_ = new_alloc_size;
   }
 
-  return write(data, offset, nbytes);
-}
-
-void WriterTile::swap(WriterTile& tile) {
-  TileBase::swap(tile);
-  std::swap(filtered_buffer_, tile.filtered_buffer_);
+  write(data, offset, nbytes);
 }
 
 /* ********************************* */
@@ -332,7 +309,7 @@ uint64_t Tile::load_chunk_data(
   }
 
   if (total_orig_size != expected_original_size) {
-    throw TileStatusException("Incorrect unfiltered tile size allocated.");
+    throw TileException("Incorrect unfiltered tile size allocated.");
   }
 
   return total_orig_size;
